# Advanced Namespace Tools blog 

### 4 January 2020

## Ultrafilters and the Language of the Waves

This post is first in a series attempting to present a coherent semi-technical summary of the research I did in the second half of 2019, focused on infinitary mathematical structures. I will describe motivations, findings, and speculations. The research process itself is documented and captured in numerous artifacts contained within the latest ANTS iso release.

- [64-bit 9ants5.64.iso.gz](//files.9gridchan.org/9ants5.64.iso.gz)

Some relevant files are /lib/CubeRead.agda and /lib/EnlightenmentEscalator.agda, along with the other Cube.agda files in /sys/games/agdacode. The gridchatlogs from 030 onward also contain a great deal of material in completely disorganized conversational format.

### Motivations 1 : Smullyan, Woodin, Hofstadter, Penrose, Gita

The logician and puzzle-crafter Raymond Smullyan was also a philosopher, and his writing dances around an implied theme that I do not recall him stating explicitly - the parallel between the self-referentiality of Godelian proofs and the concept of enlightenment in traditions such as zen buddhism. Hofstadter's GEB situates the locus of self and consciousness and identity in a "Godel vortex where all levels cross". Penrose argues that the ability of the human mind to move between and 'look from above' at systems of mathematical truth implies that conscious mathematical thought must make use of a physical embodiment which encodes mathematics which exceeds the boundaries of finite computability. The Bhagavad-Gita includes a passage where Krishna reveals the divine form to Arjuna and it is described as a kind of monstrous transcendental biologically themed fractal of eyes and limbs.

All of these ideas are deeply entangled with infinity and self-containing recursions. The branch of mathematics which has studied these things in the greatest depth and precision is set theory and model theory. Hugh Woodin's "Ultimate L" program is a good contemporary example of research with both formal precision and philosophical import. I believe that an attempt to understand the strange loops and self-containments of Hofstadter, the analogy between diagonalization and enlightenment in Smullyan, and the hypothesis of mathematical thought beyond computability in Penrose, leads naturally to the study of set theoretical ideas of large cardinal axioms, forcing principles, embeddings, and inner models, to find conceptual tools to both make these ideas more precise, and to verify or refute them.

### Motivations 2 : Free will, naming and narrative, fiction and reality

The philosophical puzzle of free will needs no introduction. In my own personal life experience, the power of of my free will to affect my own story seems as directly real and tangible as kicking a rock. In 2013 during the Harmonic Time Bind Ritual Symphony, I developed a theory of free will as a mechanism in which reality was shaped into a coherent narrative by the process of giving names and story meanings to a selection of the random events that occur in life.

I felt that the imagination was able to explore alternate/parallel universes, and then make perceptual symbol-binds and meaningful choices to select a particular branch of the multiverse to experience. The interior self was conceptualized as an inner copy of the whole modeled on the ideas that the external universe is an atom in the body of a member of a divine pantheon, and there is a sub-atom of pure divine perfection within us that reflects the totality, and this infinitary reflection between the micro-soul-god and the macro-universe-god creates an infinite regress of mirrorings, containing the unique individual free will as a dynamic emergent reflection of its unique position within the totality and degrees of freedom available for variation.

The problem is, that sounds kind of nice in combination with a sufficient dose of weed and mushrooms, but its probably not going to convince anyone who is not high as fuck. Can such ideas be given a mathematical and scientifically meaningful form, rather than being simply the ravings of a Gene Ray "Timecube" crank, or the poetic yet opaque word soup of the Deleuze/Guattari school, and make sense to people in an entirely sober and rational and objective frame of evaluation? Writing computer code that typechecks/compiles/runs and does things in the real world is one way to ground ideas in something with more tangibility and possible real-world utility than pure philosophical speculation usually possesses.

### Motivations 3 : The Language of the Waves

When Maevele and I first started our creative partnership almost seven years ago, she told me that she had spent a recent night contemplating the lake and trying to learn to understand the language of the waves. That phrase possessed an uncanny resonance for me - some deep meaning that I was at a loss to fully articulate. The Navier-Stokes equations of fluid dynamics which are not fully understood mathematically, and the quantum waveforms that are the substrate of physicality, are both in play when we watch the pattern of light reflecting off the surface of a lake. Is there an insight-producing metaphorical relationship between wave patterns and the syntax and semantics of languages?

### Seemingly Impossible Searches and Binary Hypercubes

In December 2018 I encountered two seed crystals that triggered the transformative processes of the next year. Martin Escardo's code for "Seemingly Impossible Functional Programs" and the existence of the "nCUBE Transit" variant of Plan 9, which ran on computers built around binary hypercubic interconnect topology. The fact that both of these mathematical concepts can be expressed in terms of arbitrary length vectors of binary digits (011100111...) meant that pursuing understanding of them and research into connected topics had a certain obvious harmony which developed into quite significant mathematical connections upon study. 

### Onward to Ultrafilters

So, as the Hypercubic Time Warp All-go-rhythmic Synchrony began in 2019, the following themes and ideas were already in play:

- Infinitary/recursive possibly non-computable structures as generative of consciousness and identity
- Diagonalization and self-containment as metaphors for enlightenment and the relation of soul and divine
- The idea of free will as a subset-selection principle transforming entropy into narrative structure
- The speculation that the rich theory of the higher infinite in set theory might provide models of these ideas

Particular sites of investigation and development of these ideas in 2019 were to be Plan 9 namespaces, binary hypercubes, and using binary sequences as functions from Nat -> Bool to explore mathematical ideas in a form suitable for computer code, first in Haskell, later in Agda. I was aware of ultrafilters and their importance in set theory and model theory, but I did not have nearly enough technical training to understand the research papers I would stare at on arXiv.

That changed once I met metmetmet. But that's a story for the next post, where actual computer code will re-enter the narrative.

