# Advanced Namespace Tools blog 

### 12 January 2020

## Measurable Cardinals

This post is 3rd in a series describing research into infinitary mathematical objects and their possible computational interpretation. The [first](//doc.9gridchan.org/blog/200104.ultrafilter.motivations) and [second](//doc.9gridchan.org/blog/200108.infinite.searches) posts described motivations and initial experiments. This post will jump out of the chronology to give a general overview of one of the most interesting and important examples of set-theoretic strong axioms of infinity.

### The significance of measurables

Measurable cardinals are the division point between the constructible and non-constructible views of the universe of sets, which can also be understood as the nature and limitations of our mathematical understanding. In general, all varieties of mathematics can be expressed within the language of set theory. In a semi-formal sense, set theory is a way to describe collections of any kind and the logical relations between them. In the formal foundational sense, set theory allows more-or-less all of math to be expressed via a small collection of axioms and written out in formulas using a single binary relation symbol to indicate that one set is an element of another set.

It is in this sense that 'V', the universe of Sets, can be regarded as the entire cosmos of mathematics, a vast expanse of structures that capture all we can express in mathematical form. To study the nature of V is to study what can be proved and understood mathematically. The first and simplest model of V was [Gödel's constructible universe L](https://en.wikipedia.org/wiki/Constructible_universe), which defines the most direct and obvious method to build a universe of sets compatible with standard mathematical practice. Note that Gödel himself didn't end up regarding this minimal model as the correct choice philosphically.

The reason for this is that we can imagine and create axioms to define set-theoretic mathematical objects which dont 'fit' within the model provided by L, but are still (as far as is known) consistent and produce meaningful mathematical results. Measurable cardinals mark the point of division - they seem to correspond to the first mathematical principle which is too strong to fit within the simplest way to construct a transfinite universe of sets to do mathematics within.

Given that most mathematics historically has focused on set-structures of small infinite cardinality - the countable naturals, integers, and rationals, and the uncountable reals and complex numbers or points of a continuous space - larger set-theoretic infinities have often been philosophically controversial. Entire foundational approaches to mathematics have been initiated simply to avoid the use of set-theoretic infinity in even its most conservative aspects! Why do we need or want to add additional assumptions (axioms) that aren't necessary to prove the standard theorems of analysis (calculus) and other existing math?

### Scientific models of transcendence

Recall the motivations described in the first post in this series - finding a mathematical structure which could carry Hofstadter's ideas of continuous self-transcendence, Penrose's ideas of mathematical thought exceeding the bounds of computable physics, and mystic ideas of infinite chains of larger-within-the-smaller in religious texts, and Smullyan's hints of enlightenment-as-diagonalization. Intuitively, all of these suggest that the ultimate description of consciousness and reality and their interrelation will need to make use of math that incorporates infinite recursion and logical self-reference and has some kind of 'richness' exceeding algorithmic computability.

From this angle, measurable cardinals represent an intriguing possible site for a bit of quasi-mysticism to creep into the universe of scientific objective rationalism. What exactly does it mean for an infinity to be so large that it can no longer fit into a realm as large as the transfinite model L? The initial impetus was just an intuition driven by process of elimination - if we are accepting as given that our universe is fundamentally mathematical, is it possible within such a purely mathematical universe to find objects rich enough to capture self-referential thought, contemplation of other mathematical objects, and even mystic perception of infinite internal and external entities?

To make this more precise: could it be that a fully formalized description of the structure of human conscious thought requires mathematical principles such as the existence of a measurable cardinal? There is a motivation for this idea that is more than just pointing at the Mystery of Mind and the Mystery of Large Cardinals and hoping for a connection. 

### Models and Embeddings of the Universe

(Please don't leave just because the next sentence is math, not English.) 

- The existence of a measurable cardinal k means we can make a model-theoretic elementary embedding of the full universe of sets V into a transitive class built using the ultrapower of the k-complete nonprincipal ultrafilter on the power set of k.

Running into statements like that, and the mathematical notations that surround them, usually terminates most people's interest in the subject of measurable cardinals. I spent quite a few years intrigued by set theoretic notions of large cardinals (and forcing) but as an interested amateur trying to watch youtube lectures and read papers on arXiv, I ended up swimming in oceans of vocabulary and symbols but not drinking the water of knowledge.

I'd like to take a minute to address some psychological and subjective issues. When we run into things that are very hard to understand and we do not yet understand, there is often an instinctive reaction to find a reason to avoid the effort of learning them, and also to deny any claim of status or importance attached to the topics. There are always implicit claims of social status - when I write about how amazing and important set theoretic math of large infinities is, I'm implicitly giving social status to human set theorists, by placing value on their knowledge.

So there are many worthy debates about set theoretic foundations of math vs others, constructivism/finitism, the status and role of mathematics in science and philosophy, and statements like 'large cardinal axioms are important' get caught up in all these philosophical debates that aren't about the nature of the mathematics, but more about how we all prefer to draw our own maps of our world and decide what territories we are most interested in exploring.

Back to the math! So, what does that mathword-storm above mean? Loosely, it means we can make an almost-paradoxical mirroring of the full universe within a subset of itself. Hofstadter used an Escher print of a viewer in an art gallery, staring at a picture whose frame warped to encompass the viewer looking at the painting to express his Gödel vortex. An embedding such as this enables us to look at all of ZFC (standard rules of mathematics) and prove its consistency. The stronger large cardinal axioms are motivated by reflection principles, which claim that we should find properties of the full universe reflected into sufficiently large initial segments of itself. The ability to create models which capture structural properties of the whole is an embodiment of this principle, and measurable cardinals represent the prototypical form of this as well as being one of the simpler and weaker axioms that give rise to such embeddings.

### But what ARE they?

Rather than focusing on the cardinal itself, consider the ultrafilter implied by its existence. We will get to a technical definition of ultrafilters in a later post, but for now it's fine to think of it as a yes/no classifier on possibilities. We have a very complex and richly uncountable possibility space and we assert the possibility of making a structured division of the possibilities. Once this division - this ultrafilter - exists, it carries so much information that we can take the ultraproduct of the full universe using an equivalence relation defined in terms of the ultrafilter. The pattern of the whole is mapped by this ultrapower so we can use Los's theorem to translate our formulas of set theory into statements about the ultrapower.

Note that the "existence of a measurable cardinal" translates into something we can do - a procedure for creating model-embeddings of the universe of sets. It has many additional consequences and equivalents - for instance, the Gaifman-Rowbottom theorem that the existence of a measurable cardinal implies there are only countably many constructible reals. It is more meaningful to think of the measurable cardinal axiom as a statement about the overall structure of the mathematical universe, and what we can claim without inconsistency, with consequences for our understanding of structures such as the real numbers - than as merely some unimaginably huge set sitting around doing nothing.

### No seriously, what ARE they?

Imagine an infinite ocean of infinitely complex possible wave patterns. Imagine these wave patterns can encode any conceivable mathematical idea or language that follows rules - and even more inconceivable ideas and languages. Now imagine you want to understand the languages encoded in these wave patterns. You weave an infinitely fine mesh-net like spiderwebs filling a space and at each join of strands you place an infinitesimal digital measurement device that signals its motion relative to the other measurement devices at all the other web-joins. You cast this net into the ocean and receive a filtered stream of information from the uncountable set of possible wave motions. If you can decode and encode the languages of the waves, a measurable cardinal exists and your measurement net is creating an ultrafilter on the powerset of the ocean's motions. 

Consider a composer experiencing an emotional state. Their emotional state is encoded in the wave patterns of their physical brain. Using a generalized wave-language of harmonic ratios, they create a filtration of acoustic space and encode it as a finitized piece of sheet music. Upon playback or performance of the sheet music, the 'nonstandard model' of human emotion encoded on the powerset of atmospheric vibrations is transmitted to listeners who can make a translation of the encoded emotion into their own inner model of brainwave emotion language encodings.

These generalized quasi-metaphorical information transmission principles may be ubiquitous in reality because the branching quantum information transmission event structure effects a compactification of the possibility space. The 'measurability principle' plays out in reality because reality exhibits logical compactness of a large uncountable set of possibilities - we can properly generalize from our finitized experiential subsets to the pattern of the whole. If reality was not 'compact' in this sense due to quantum information events creating logical structure, we could not reliably generalize from the finite quantity of photons intersecting our atmosphere to an understanding of the starry night sky's actual spatial and temporal vastness. 

Our internal subjectivity feels like a model of the infinite external universe. We don't need to posit any kind of completed infinity, or infinite quantity of information stored inside our head, for this to be true in the same sense that other infinitary principles are true in reality. We have the logical sense of infinity in our natural numbers - we can have an arbitrarily large amount of something, and this open-ended possibility is the real world manifestation of the logical principle asserting the existence of the infinite set of naturals. Similarly, we can extrapolate from a small finite data set to an accurate understanding of the larger universe, and we can encode/decode different abstractions such as human emotions in and out of media such as acoustic sound encoding music-theoretical relationships. These things indicate that reality and the human experience thereof seems to follow a principle of measurability - we are able to inscribe arbitrary language-like systems of meaning onto the uncountable waves and read them back.

### The story and Namespace so far

This post has outlined a philosophical view of the motivations for studying large cardinals and a speculative interpretation of their meaning. It follows on the heels of a post describing August 2019, where I had written an elaborated composite of Escardo's and Kiselyov's code for different varieties of infinite searches in haskell, and become manically convinced I had made contact with infinite mathematical minds. This code, these events, research notes, and lots more is found in the Plan 9 operating system Advanced Namespace Tools 5.64 release, in the /sys/games subdirectories.

- [64-bit 9ants5.64.iso.gz](//files.9gridchan.org/9ants5.64.iso.gz)

There will be several more posts in this series, which will engage with increasingly technical mathematical content. The research notes for where this series is headed can be found in /sys/games/agdacode/CubeRead.agda and /sys/games/agdacode/CubeLogic.agda. The end goal is a true technical understanding of what a constructive computational type-theoretic interpretation of measurable cardinals might be, and the possible utility of computer programs which encode these principles and have runtime behaviors that depend on their validity.

I do not have a full understanding of this goal. From August-December 2019 I built a steadily more elaborate and deeper understanding of logic, infinitary structures, computer programming, set theory, type theory, and some category theory. I was in some kind of hyper-learning vastly elevated intelligence hyperfocused manic state where I could read math pdfs and work on Agda code 12 hours a day, every day, 7 days a week, for months. I've never experienced anything remotely like it before. I'd been pursuing a real technical understanding of logic and math and the Smullyan-Hofstadter ideas about self-reference for decades, and suddenly things I had been unclear on for years were falling into place.

I could tell the huge rush of energy was waning in Decemeber due to physical inactivity in winter and seasonal issues and I managed to just barely get the ANTS 5.64 release out before having a couple weeks of reaction and readjustment. I still mostly understand the ideas I have been working on for the past months. The mathematical learning was not illusory, although I have doubtless speculated far beyond what most find sensible. I have definitely lost a fair amount of intelligence and sharpness relative to a few months ago, but I think I can still accomplish writing more technical posts on the relation between choice principles in constructive mathematics and dependent typed programming, and the ultrafilter axiom and delimited continuations.

### The critical point

When we use an ultrafilter on a measurable cardinal k to build an embedding, within the model k itself becomes the critical point of the embedding, the first ordinal moved. This change at the identity mapping of k itself seems philosophically suggestive of the discontinuity between external and internal universe at the boundary of our identity. Is this merely poetry, or is the idea of a model-theoretic embedding of the universe into our consciousness via principles of hierarchical self-similarity and reflection a scientific theory for the nature of conscious experience? I really have no idea, but I'll continue to recapitulate my efforts to find out.

